/* ======================================================================= */
/* cell_sim.cpp                                                            */
/*    This file contains the abstraction of Cell Simulator.                */
/*    The Cell Simulator (ViruSIM) is a single system-wide object          */
/*    Abstractions for external input to the system are also present here  */
/* ======================================================================= */

#include "basic.h"
#include "../threads/thread.h"
#include "display.h"

# if 0
void simDisplay()
{
	int i;
	for(i=0; i<nelements; i++)
		printf("%3d",cells[i].id);
	printf("\n");
	for(i=0; i<nelements; i++)
		printf("%3d",IsCellActive[i]);
	printf("\n");
	for(i=0; i<nelements; i++)
		printf("%3d",(int)cells[i].cell_state);
	printf("\n\n");
}
# endif

SimCell *CellSimulator::getNewCell ()
{
  if (nspace == 0) 
    {
      if (!growCells ()) 
        {
          printf("Cell Simulator Error: Not enough cells available to proceed\n");
          exitProgram ();
        }
    }
  
  assert (nspace > 0);
  nspace--;

  SimCell *new_cell = new SimCell (g_cellProgram, nelements);
  new_cell->init ();

  cells[nelements] = new_cell;
  nelements++;
  return new_cell;
}

void CellSimulator::freeCell (int index)
{
  cells[index]->cell_state = CELL_NULL_STATE;
}

/*void  NodeMessage::init(int _src, int _from, int _type, int vNumber)
{
		src=_src;
		from=_from;
		type = _type;
		virusNumber = vNumber;
}*/

bool CellSimulator::growCells ()
{
  int i;
  SimCell **newelements; 
  size_t newspace = nelements * 2;

  newelements = new SimCell *[newspace];

  if (newelements == NULL) {
    newspace = nelements + CELLS_BASE_SIZE;
    newelements = new SimCell *[newspace];
  }

  if (newelements == NULL) {
    return false;
  }

  nspace = newspace - nelements;

  for (i = 0; i < nelements; i++)
    {
      newelements[i] = cells[i];
    }
  
  delete cells;
  cells = newelements;
  return true;
}

void CellSimulator::init (bool p_batch, char *simFile)
{
  batch = p_batch;

  minX = minY = -1;
  maxX = maxY = +1;
  w_width = W_WIDTH;
  w_height = W_HEIGHT;
  
  comm_radius = .1f;
  
  nspace = CELLS_BASE_SIZE;
  nelements = 0;
  cells = new SimCell *[nspace];
  
  if (simFile == NULL)
    {
      /* Setup the one Initial Cell at (0,0,0) */
      SimCell *initcell = getNewCell ();
      initcell->setValues (0, 0, 0, INITIAL_CELL_RADIUS, M_PI_2, 0);
      initcell->cell_state = g_cellProgram->initialState;
      
      instructions = NULL;
    }
  else
    {
      cerr << "Loading simulator file: " << simFile << endl;
      
      FILE *f;
      
      if ((f = fopen(simFile,"r+")) == NULL)
        {
          cerr << "Cell Simulator Error: Unable to open sim file " << simFile << endl;
          exit (2);
        }
      
      instructions = new SimInstructions ();
      instructions->loadInstructions (f);
    }
  
  interpretSimulationInstructions ();
}

void CellSimulator::interpretSimulationInstructions()
{
  if (instructions) {
    instructions->interpretInstructions (this);
  }
}

void simulator_thread (int dummy)
{
  Chemical null_chem;
  int i;
  
  g_tick++;
  // YIELD;

  while (1)
    {
		null_chem.id = 0;
		null_chem.concentration = 0;
      
		for (i = 0; i < g_simulator.numCells (); i++)
        {
          g_simulator.getCell (i)->mbox.send(&null_chem,sizeof(Chemical));
        }
      

		for (i = 0; i < g_simulator.numCells (); i++)
        {
          SimCell *cell = g_simulator.getCell (i);

          if (cell->isNewCell)
            {
              cell->isNewCell = false;
            }
          else 
            {
              if (cell->cell_state == g_cellProgram->finalState)
                {
                  ;
                }
              else if (cell->cell_state != CELL_NULL_STATE)
                {
                  cell->runCellProgram ();
                }
            }
        }

      g_simulator.showStateSummary ();
		g_simulator.writeStateDetails ();
  
      if (!g_simulator.isBatch ()) {
        YIELD; // display thread
      }

		g_tick++;
		g_simulator.interpretSimulationInstructions ();
		g_simulator.writeStateDetails ();
    }
}

void CellSimulator::disperseChemical (int node_id, Emission *emit)
{
  float x1,y1,x2,y2,r;
  int i;

  x1 = getCell (node_id)->x;
  y1 = getCell (node_id)->y;

  ChemicalId chemid = emit->getChemical ();
  float conc = emit->getConcentration ();

  for (i = 0; i < nelements; i++)
    {
      x2 = getCell (i)->x;
      y2 = getCell (i)->y;
      
      r = sqrt ((x1-x2)*(x1-x2) + (y1-y2)*(y1-y2));

      Chemical chem;
      chem.id = chemid;
      chem.concentration = conc * (exp (-r/4));
      getCell (i)->mbox.send(&chem, sizeof(Chemical));
    }
}

float CellSimulator::rTransX (int x)
{
	return (2*(float)x/w_width)-1;
}

float CellSimulator::rTransY (int y)
{
	return 1-(2*(float)y/w_height);
}

int CellSimulator::transX(float x)
{
	return (int)( (w_width/2)*(1+x) );
}

int CellSimulator::transY(float y)
{
	return (int) ((w_height/2)*(1-y) );
}

SimCell *CellSimulator::getCell (int i) const
{
  assert (i >= 0 && i < nelements);
  return cells[i];
}

int CellSimulator::numCells () const
{
  return nelements;
}

CellStateId CellSimulator::getState (int i)
{
  return getCell (i)->cell_state;
}

int CellSimulator::getX (int i)
{
	return (int) ((w_width/2) * (getCell (i)->x+1));
}

int CellSimulator::getY (int i)
{
	return (int) ((w_height/2)* (1 - getCell (i)->y) );
}

int CellSimulator::getR (int i)
{
	return (int) ((w_height/2) * (getCell (i)->r));
}

void CellSimulator::showStateSummary ()
{
  cerr << "Step " << g_tick << ": " << numCells () << " cells.\n";
}

void CellSimulator::writeStateDetails ()
{
	g_DumpFile << g_tick << " " << numCells() << endl;
	for (int i = 0; i < numCells(); i++)
	{
		double x, y, z;
		int r, g, b, color;
		if(getCell(i)->isAlive())
		{
		getCell(i)->getCoordinates(x,y,z);
		color = getCell(i)->getDisplayColor();
		r = (color & 0xFF);
		g = (color >> 8) & 0xFF;
		b = (color >> 16) & 0xFF;
		g_DumpFile << x << " " << y << " " << z << " " << getCell(i)->r << " " << r << " " << g << " " << b << endl;
		}
	}
}

