/* ======================================================================= */
/* SimCell.cpp                                                             */
/*    This file contains the abstraction of the Cell in the VirusSIM.      */
/*    A cell contains a VirusTable                                         */
/* ======================================================================= */

# include "basic.h"

SimCell::SimCell(CellProgram *cpa, int cell_id)
{
  //num_neighbours = 0;
  cell_program = cpa;
  cell_state = CELL_NULL_STATE;
  isNewCell = true;
  id = cell_id;
}

SimCell::SimCell(CellProgram *cpa, int cell_id, double px, double py, 
                 double pz, double pr, double pAxis, double pNormal) 
{
  //! how can I use 1-param constructor here?
  cell_program = cpa;
  id = cell_id;
  cell_state = CELL_NULL_STATE;
  isNewCell = true;
  setValues (px, py, pz, pr, pAxis, pNormal);
  cerr << "Created cell: " << cell_id << endl;
}

void SimCell::init()
{
  Chemical temp;
  
  while (mbox.recv (&temp, sizeof(Chemical)) >= 0)
    {
      ;
    }
}

void SimCell::setValues(double px, double py, double pz, double pr, double pAxis, double pNormal)
{
  x = px;
  y = py;
  z = pz;
  r = pr;
  abAxis = pAxis;
  abNormal = pNormal;
}

void SimCell::moveBy(float dx, float dy, float dz)
{
  x += dx;
  y += dy;
  z += dz;
}

void SimCell::getCoordinates(double &px, double &py, double &pz)
{
	px = x;
	py = y;
	pz = z;
}

static double square (double x) { return x * x; }

double SimCell::getDistanceFrom (double p_x, double p_y, double p_z)
{
  return (sqrt (square (x - p_x) + square (y - p_y) + square (z - p_z)));
}

int SimCell::getDisplayColor ()
{
  if (cell_program->hasStateColor (cell_state))
    {
      return cell_program->getStateColor (cell_state);
    }
  else
    {
      int s = (int) cell_state;
      int r = 255 - (s % 10) * 25;
      int g = 255 - ((s / 4) % 10) * 25;
      int b = (s % 40) * 6;

      return (int) ((b << 16) + (g << 8) + (r));
    }
}



/*void Cell::addNeighbour(Cell *ptr, double distance, double angle)
{
	neighbour_ptr[num_neighbours]=ptr;
	neighbour_distance[num_neighbours]=distance;
	neighbour_angle[num_neighbours]=angle;
	num_neighbours++;
}

void Cell::removeNeighbour(int index)
{
		if(num_neighbours < 1)
			return;
		for(int i=index+1; i<num_neighbours; i++)
		{
			neighbour_ptr[i-1]=neighbour_ptr[i];
			neighbour_distance[i-1]=neighbour_distance[i];
			neighbour_angle[i-1]=neighbour_angle[i];
		}
		num_neighbours--;
}
*/

void SimCell::splitUsingNormal_1 (SimCell *daughter)
{
	/*double dist = ((sqrt(2)-1)/(sqrt(2)))*r;*/
	double dist = r;
	double axis1, axis2, normal1, normal2, temp;

	axis1 = normalAngle(abAxis - M_PI/4.0);
	axis2 = normalAngle(abAxis + M_PI/4.0);
	normal1 = normalAngle(abNormal - M_PI/4.0);
	normal2 = normalAngle(abNormal + M_PI/4.0);
	if(!approximatelySame(axis1-M_PI_2,normal1))
	{
		temp = axis1; axis1 = axis2; axis2 = temp;
		temp = normal1; normal1 = normal2; normal2 = temp;
	}

	daughter->setValues (x - dist/2 * cos(abNormal), 
                        y - dist/2 * sin(abNormal),
                        z, pow(r*r*r / 2,0.333333), axis2, normal2);

	x = x + dist/2 * cos(abNormal);
	y = y + dist/2 * sin(abNormal);
	r = pow(r*r*r / 2,0.333333);
	abAxis = axis1;
	abNormal = 	normal1;

	/*for(i=0; i<num_neighbours; i++)
		daughter->addNeighbour(neighbour_ptr[i],neighbour_distance[i],neighbour_angle[i]);
	addNeighbour(&daughter,dist/2,0);
	daughter->addNeighbour(this,dist/2,0);*/
}

void SimCell::splitUsingNormal_2 (SimCell *daughter)
{
	/*double dist = ((sqrt(2)-1)/(sqrt(2)))*r;*/
	double dist = r;
	double axis1, axis2, normal1, normal2, temp;

	axis1 = normalAngle(abAxis - M_PI/4.0);
	axis2 = normalAngle(abAxis + M_PI/4.0);
	normal1 = normalAngle(abNormal - M_PI/4.0);
	normal2 = normalAngle(abNormal + M_PI/4.0);
	if(!approximatelySame(axis1-M_PI_2,normal1))
	{
		temp = axis1; axis1 = axis2; axis2 = temp;
		temp = normal1; normal1 = normal2; normal2 = temp;
	}

	daughter->setValues (x, 
                        y,
                        z - dist/2, pow(r*r*r / 2,0.333333), axis2, normal2);

	z += dist/2;
	r = pow(r*r*r / 2,0.333333);
	abAxis = axis1;
	abNormal = 	normal1;

	/*for(i=0; i<num_neighbours; i++)
		daughter->addNeighbour(neighbour_ptr[i],neighbour_distance[i],neighbour_angle[i]);
	addNeighbour(&daughter,dist/2,0);
	daughter->addNeighbour(this,dist/2,0);*/
}

void SimCell::splitUsingAxisPolarised (SimCell *daughter)
{
	/*double dist = ((sqrt(2)-1)/(sqrt(2)))*r;*/
	double dist = r;

	daughter->setValues (x + dist/2 * cos(abAxis), 
                        y + dist/2 * sin(abAxis), z, pow(r*r*r / 2,0.333333), abAxis, normalAngle(-abNormal));
	x = x - dist/2 * cos(abAxis);
	y = y - dist/2 * sin(abAxis);
	r = pow(r*r*r / 2,0.333333);
	/*for(i=0; i<num_neighbours; i++)
		daughter->addNeighbour(neighbour_ptr[i],neighbour_distance[i],neighbour_angle[i]);
	addNeighbour(&daughter,dist/2,0);
	daughter->addNeighbour(this,dist/2,0);*/
}

void SimCell::splitUsingAxisUnpolarised (SimCell *daughter)
{
	/*double dist = ((sqrt(2)-1)/(sqrt(2)))*r;*/
	double dist = r;
	double axis2, normal2;

	axis2 = normalAngle(abAxis - M_PI);
	normal2 = normalAngle(abNormal - M_PI);
	daughter->setValues (x - dist/2 * cos(abAxis), y - dist/2 * sin(abAxis), z, pow(r*r*r / 2,0.333333), axis2, normal2);
	x = x + dist/2 * cos(abAxis);
	y = y + dist/2 * sin(abAxis);
	r = pow(r*r*r / 2,0.333333);
	/*for(i=0; i<num_neighbours; i++)
		daughter->addNeighbour(neighbour_ptr[i],neighbour_distance[i],neighbour_angle[i]);
	addNeighbour(&daughter,dist,0);
	daughter->addNeighbour(this,dist,0);*/
}

void SimCell::runCellProgram()
{
  SimCell *new_cell;
  CellStateId out1, out2;
  t_splitKind mode;
  Chemical chem;

  /*
  ** Secrete chemicals
  */

  EmissionList *emissions = cell_program->getEmissions (cell_state);
  
  emissions->startIterator ();

  while (emissions->hasMore ()) {
    Emission *emit = emissions->getNext ();
    g_simulator.disperseChemical (id, emit);
  }

  emissions->finishIterator ();

  Environment *env = new Environment ();
  
  while (mbox.recv (&chem, sizeof (Chemical)) >= 0)
    {
      env->reflect (&chem);
    }

  cell_program->getTransition (cell_state, env, &out1, &out2, &mode);

  // cerr << "Transition: " << cell_state << " ==> " << out1 << ", " << out2 << "\n";

  if (out2 == CELL_NULL_STATE)
    {
		cell_state = out1;

		if (out1 == CELL_NULL_STATE) 
        {
          g_simulator.freeCell(id);
        }

		return;
	}

	new_cell = g_simulator.getNewCell();

	if (new_cell == NULL)
     {
       return;
     }

   switch (mode) {
   case SPLIT_AXIS:
     splitUsingAxisUnpolarised (new_cell);
     break;
   case SPLIT_NORMAL_1:
     splitUsingNormal_1 (new_cell);
     break;
   case SPLIT_NORMAL_2:
     splitUsingNormal_2 (new_cell);
     break;
   case SPLIT_POLARIZED:
     splitUsingAxisPolarised (new_cell);
     break;
   default:
     cerr << "Bad mode for split!\n";
     assert (false);
	}

	new_cell->cell_state = out2;
	cell_state = out1;
}

double normalAngle(double A)
{
	if((A > M_PI)||(A <= -M_PI))
		A -= 2*M_PI*(int)( A / (2*M_PI) );
	if(A == -M_PI)
		A = -A;

	return A;
}

double positiveAngle(double A)
{
	if((A>=2*M_PI) || (A<0))
		A -= 2*M_PI*(int)( A / (2*M_PI) );
	if(A<0)
		A -= 2*M_PI*((int)( A / (2*M_PI) ) - 1);

	return A;
}

int approximatelySame(double A, double B)
{
	if(positiveAngle(A-B) <= 0.01)
		return 1;
	else if(positiveAngle(A-B) >= (M_PI * 2 - 0.01))
		return 1;
	else
		return 0;
}

string SimCell::destructiveToString ()
{
  char buf[BUFSIZE];
  string res;

  snprintf (buf, BUFSIZE, "%3d (%f %f %f) : ", 
            (int) cell_state, x, y, z);
  res = buf;

  Chemical chem;
		
  while ((mbox.recv (&chem, sizeof(Chemical))) >= 0) {
    snprintf (buf, BUFSIZE, "(%d %f)", (int) chem.id, chem.concentration);
    res += buf;
  }

  return res;
}

string SimCell::toString ()
{
  char buf[BUFSIZE];
  string res;

  snprintf (buf, BUFSIZE, "%3d (%f %f %f) %f ", 
            (int) cell_state, x, y, z, r);
  res = buf;
  return res;
}












